#include <linux/types.h>

/**
 * struct pako_symbol - Pako's analogue of an ELF symbol
 * @name:		The ELF name of the symbol
 * @label:		A unique Pako name for the symbol
 * @candidate_vals:	A linked list of possible values for the symbol, or NULL
 * @value:		The value of the symbol (valid when vals is NULL)
 **/
struct pako_symbol {
	const char *name;
	const char *label;
/* private: */
	struct list_head *candidate_vals;
	unsigned long value;
};

/**
 * struct pako_reloc - Pako's analogue of an ELF relocation
 * @blank_addr:		The address of the relocation's storage unit
 * @symbol:		The pako_symbol associated with this relocation
 * @howto:		The information regarding the relocation type
 * @insn_addend:	The part of the ELF addend resulting from quirks of
 *			the instruction one of whose operands is the relocation.
 *			For example, this is -4 on x86 pc-relative jumps.
 * @target_addend:	The rest of the ELF addend.  This is equal to the offset
 *			against the symbol that the relocation refers to.
 **/
struct pako_reloc {
	unsigned long blank_addr;
	struct pako_symbol *symbol;
	const struct pako_reloc_howto *howto;
	long insn_addend;
	long target_addend;
};

enum pako_reloc_howto_type {
	PAKO_HOWTO_RELOC,
	PAKO_HOWTO_RELOC_PATCH,
	PAKO_HOWTO_DATE,
	PAKO_HOWTO_TIME,
	PAKO_HOWTO_BUG,
	PAKO_HOWTO_EXTABLE,
	PAKO_HOWTO_SYMBOL,
};

/**
 * struct pako_reloc_howto - Pako's relocation type information
 * @type:		The type of the relocation
 * @pcrel:		Is the relocation PC relative?
 * @size:		The size, in bytes, of the item to be relocated
 * @dst_mask:		Bitmask for which parts of the instruction or data are
 * 			replaced with the relocated value
 * 			(based on dst_mask from GNU BFD's reloc_howto_struct)
 * @rightshift:		The value the final relocation is shifted right by;
 * 			used to drop unwanted data from the relocation
 * 			(based on rightshift from GNU BFD's reloc_howto_struct)
 * @signed_addend:	Should the addend be interpreted as a signed value?
 **/
struct pako_reloc_howto {
	enum pako_reloc_howto_type type;
	int pcrel;
	int size;
	long dst_mask;
	unsigned int rightshift;
	int signed_addend;
};

#if BITS_PER_LONG == 32
#define PAKO_CANARY 0x77777777UL
#elif BITS_PER_LONG == 64
#define PAKO_CANARY 0x7777777777777777UL
#endif /* BITS_PER_LONG */

/**
 * struct pako_section - Pako's analogue of an ELF section
 * @symbol:		The pako_symbol associated with this section
 * @size:		The length, in bytes, of this section
 * @address:		The address of the section
 * @flags:		Flags indicating the type of the section, whether or
 *			not it has been matched, etc.
 **/
struct pako_section {
	struct pako_symbol *symbol;
	unsigned long address;
	unsigned long size;
	unsigned int flags;
	const unsigned char **match_map;
	int unmatched;
};
#define PAKO_SECTION_TEXT 0x00000001
#define PAKO_SECTION_RODATA 0x00000002
#define PAKO_SECTION_DATA 0x00000004
#define PAKO_SECTION_STRING 0x00000008
#define PAKO_SECTION_MATCH_DATA_EARLY 0x00000100
#define PAKO_SECTION_MATCHED 0x10000000

#define MAX_TRAMPOLINE_SIZE 5

enum pako_patch_type {
	PAKO_PATCH_TEXT,
	PAKO_PATCH_DATA,
	PAKO_PATCH_EXPORT,
};

/**
 * struct pako_patch - A replacement that Pako should perform
 * @oldaddr:		The address of the obsolete function or structure
 * @repladdr:		The address of the replacement function
 * @type:		The type of the pako patch
 * @size:		The size of the patch
 * @contents:		The bytes to be installed at oldaddr
 * @vaddr		The address of the page mapping used to write at oldaddr
 * @saved:		The bytes originally at oldaddr which were
 * 			overwritten by the patch
 **/
struct pako_patch {
	unsigned long oldaddr;
	unsigned long repladdr;
	enum pako_patch_type type;
	unsigned int size;
	void *contents;
/* private: */
	void *vaddr;
	void *saved;
};

#ifdef PAKO_STANDALONE
struct pako_system_map {
	const char *label;
	unsigned long nr_candidates;
	const unsigned long *candidates;
};
#endif /* PAKO_STANDALONE */

#ifdef __KERNEL__
#include <linux/module.h>
#include <linux/stringify.h>
#include <linux/version.h>

#define _KS_PASTE(x, y) x##y
#define KS_PASTE(x, y) _KS_PASTE(x, y)
#define PAKO_UNIQ(s) KS_PASTE(s##_, PAKO_MID)
#define PAKO_KID_UNIQ(s) KS_PASTE(s##_, PAKO_KID)
#ifdef PAKO_STANDALONE
#define init_pako_mod_change PAKO_KID_UNIQ(init_pako_mod_change)
#define cleanup_pako_mod_change PAKO_KID_UNIQ(cleanup_pako_mod_change)
#endif

/**
 * struct pako_module_list_entry - A record of a pako_mod_change's target
 * @target_mod_name:	The name of the pako_mod_change's target module
 * @new_code_mod_name:	The name of the pako_mod_change's new_code module
 * @applied:		Whether the pako_mod_change was applied or not (this
 *			will be false for pako_mod_changes patching targets
 *			that are not loaded when the partial flag is set)
 **/
struct pako_module_list_entry {
	const char *target_mod_name;
	const char *new_code_mod_name;
	const char *kid;
	bool applied;
/* private: */
	struct list_head update_list;	/* list head for this is per-update */
	struct list_head list;	/* list head for this is global */
};

/* List of all pako modules and the module they patch */
extern struct list_head pako_modules;

/* There are two actions, apply and reverse */
#define KS_ACTIONS 2
enum pako_action {
	KS_APPLY,
	KS_REVERSE,
};

/**
 * struct pako_hooks - Hooks to be run during an action (apply or reverse)
 * @pre:			Runs before the action;
 * 				may return nonzero to abort the action
 * @check:			Runs inside stop_machine before the action;
 * 				may return nonzero to abort the action
 * @intra:			Runs inside stop_machine during the action
 * @post:			Runs after the action is successfully performed
 * @fail:			Runs if the action is aborted for any reason
 */
struct pako_hooks {
	const typeof(int (*)(void)) *pre, *pre_end, *check, *check_end;
	const typeof(void (*)(void)) *intra, *intra_end, *post, *post_end,
	    *fail, *fail_end;
};

/**
 * struct pako_code - Pako metadata for an object
 * @relocs:		The Pako relocations for the object
 * @symbols:		The Pako symbols for the object
 * @sections:		The Pako sections for the object
 **/
struct pako_code {
	struct pako_reloc *relocs, *relocs_end;
	struct pako_section *sections, *sections_end;
	struct pako_symbol *symbols, *symbols_end;
#ifdef PAKO_STANDALONE
	struct pako_system_map *system_map, *system_map_end;
#endif /* PAKO_STANDALONE */
};

/**
 * struct pako_mod_change - Data for one module modified by a Pako update
 * @name:			The name of the new_code module for the change
 * @kid:			The Pako unique identifier for the change
 * @target_name:		The name of the module modified by the change
 * @new_code_mod:		The new_code module for the change
 * @old_code:			The old code for run-pre matching
 * @new_code:			The new code to switch to
 * @patches:			The function replacements in the change
 * @patches_end:		The end pointer for patches array
 * @hooks:			Hooks to be run during apply and reverse
 * @update:			The atomic update the change is part of
 * @target:			The module modified by the change
 * @safety_records:		The ranges of addresses that must not be on a
 *				kernel stack for the patch to apply safely
 **/
struct pako_mod_change {
	const char *name;
	const char *kid;
	const char *target_name;
#ifdef PAKO_STANDALONE
	unsigned long map_printk;
#endif /* PAKO_STANDALONE */
	struct module *new_code_mod;
	struct pako_code old_code, new_code;
	struct pako_patch *patches, *patches_end;
	struct pako_hooks hooks[KS_ACTIONS];
/* private: */
	struct update *update;
	struct module *target;
	struct list_head temp_labelvals;
	struct list_head safety_records;
	struct list_head list;
};


int init_pako_mod_change(struct pako_mod_change *change);

void cleanup_pako_mod_change(struct pako_mod_change *change);

#endif /* __KERNEL__ */
